# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

import numpy as np
import torch

from qucumber.utils import cplx


class TestCplx(unittest.TestCase):
    def assertTensorsEqual(self, a, b, msg=None):
        self.assertTrue(torch.equal(a, b), msg=msg)

    def assertTensorsAlmostEqual(self, a, b, tol=1e-7, msg=None):
        self.assertTrue(((a - b).abs() <= tol).all(), msg=msg)

    def test_make_complex_vector(self):
        x = torch.tensor([1, 2, 3, 4])
        y = torch.tensor([5, 6, 7, 8])
        z = cplx.make_complex(x, y)

        expect = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8]])

        self.assertTensorsEqual(expect, z, msg="Make Complex Vector failed!")

    def test_make_complex_vector_with_zero_imaginary_part(self):
        x = torch.tensor([1, 2, 3, 4])
        z = cplx.make_complex(x)

        expect = torch.tensor([[1, 2, 3, 4], [0, 0, 0, 0]])

        self.assertTensorsEqual(
            expect, z, msg="Making a complex vector with zero imaginary part failed!"
        )

    def test_make_complex_matrix(self):
        x = torch.tensor([[1, 2], [3, 4]])
        y = torch.tensor([[5, 6], [7, 8]])
        z = cplx.make_complex(x, y)

        expect = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        self.assertTensorsEqual(expect, z, msg="Make Complex Matrix failed!")

    def test_real_part_of_vector(self):
        x = torch.tensor([1, 2])
        y = torch.tensor([5, 6])
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(x, cplx.real(z), msg="Real part of vector failed!")

    def test_imag_part_of_vector(self):
        x = torch.tensor([1, 2])
        y = torch.tensor([5, 6])
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(y, cplx.imag(z), msg="Imaginary part of vector failed!")

    def test_real_part_of_matrix(self):
        x = torch.tensor([[1, 2], [3, 4]])
        y = torch.tensor([[5, 6], [7, 8]])
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(x, cplx.real(z), msg="Real part of matrix failed!")

    def test_imag_part_of_matrix(self):
        x = torch.tensor([[1, 2], [3, 4]])
        y = torch.tensor([[5, 6], [7, 8]])
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(y, cplx.imag(z), msg="Imaginary part of matrix failed!")

    def test_real_part_of_tensor(self):
        x = torch.randn(3, 3, 3)
        y = torch.randn(3, 3, 3)
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(
            x, cplx.real(z), msg="Real part of rank-3 tensor failed!"
        )

    def test_imag_part_of_tensor(self):
        x = torch.randn(3, 3, 3)
        y = torch.randn(3, 3, 3)
        z = cplx.make_complex(x, y)

        self.assertTensorsEqual(
            y, cplx.imag(z), msg="Imaginary part of rank-3 tensor failed!"
        )

    def test_bad_complex_matrix(self):
        with self.assertRaises(RuntimeError):
            x = torch.tensor([[1, 2, 3]])
            y = torch.tensor([[4, 5, 6, 7]])
            return cplx.make_complex(x, y)

    def test_elementwise_mult(self):
        z1 = torch.tensor([[2, 3, 5], [6, 7, 2]], dtype=torch.double)
        z2 = torch.tensor([[1, 2, 2], [3, 4, 8]], dtype=torch.double)

        expect = torch.tensor([[-16, -22, -6], [12, 26, 44]], dtype=torch.double)

        self.assertTensorsEqual(
            cplx.elementwise_mult(z1, z2),
            expect,
            msg="Elementwise multiplication failed!",
        )

    def test_elementwise_div(self):
        z1 = torch.tensor([[2, 3, 5], [6, 7, 2]], dtype=torch.double)
        z2 = torch.tensor([[1, 2, 2], [3, 4, 8]], dtype=torch.double)

        expect = torch.tensor(
            [[2, (17 / 10), (13 / 34)], [0, (1 / 10), (-9 / 17)]], dtype=torch.double
        )

        self.assertTensorsAlmostEqual(
            cplx.elementwise_division(z1, z2),
            expect,
            msg="Elementwise division failed!",
        )

    def test_elementwise_div_fail(self):
        with self.assertRaises(ValueError):
            z1 = torch.tensor([[2, 3], [6, 7]], dtype=torch.double)
            z2 = torch.tensor([[1, 2, 2], [3, 4, 8]], dtype=torch.double)
            return cplx.elementwise_division(z1, z2)

    def test_scalar_vector_mult(self):
        scalar = torch.tensor([2, 3], dtype=torch.double)
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)

        expect = torch.tensor([[-7, -8], [9, 14]], dtype=torch.double)

        self.assertTensorsEqual(
            cplx.scalar_mult(scalar, vector),
            expect,
            msg="Scalar * Vector multiplication failed!",
        )

    def test_scalar_matrix_mult(self):
        scalar = torch.tensor([2, 3], dtype=torch.double)
        matrix = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)

        expect = torch.tensor(
            [[[-13, -14], [-15, -16]], [[13, 18], [23, 28]]], dtype=torch.double
        )

        self.assertTensorsEqual(
            cplx.scalar_mult(scalar, matrix),
            expect,
            msg="Scalar * Matrix multiplication failed!",
        )

    def test_scalar_mult_overwrite(self):
        scalar = torch.tensor([2, 3], dtype=torch.double)
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)
        out = torch.zeros_like(vector)
        expect = torch.tensor([[-7, -8], [9, 14]], dtype=torch.double)

        cplx.scalar_mult(scalar, vector, out=out)

        self.assertTensorsEqual(
            out,
            expect,
            msg="Scalar * Vector multiplication with 'out' parameter failed!",
        )

    def test_scalar_mult_overwrite_fail(self):
        scalar = torch.tensor([2, 3], dtype=torch.double)
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)

        with self.assertRaises(RuntimeError):
            cplx.scalar_mult(scalar, vector, out=vector)

        with self.assertRaises(RuntimeError):
            cplx.scalar_mult(scalar, vector, out=scalar)

    def test_matrix_vector_matmul(self):
        matrix = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)

        expect = torch.tensor([[-34, -42], [28, 48]], dtype=torch.double)

        self.assertTensorsEqual(
            cplx.matmul(matrix, vector),
            expect,
            msg="Matrix * Vector multiplication failed!",
        )

    def test_matrix_matrix_matmul(self):
        matrix1 = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)
        matrix2 = torch.tensor([[[1, 0], [3, 0]], [[0, 6], [0, 8]]], dtype=torch.double)
        expect = torch.tensor(
            [[[7, -78], [15, -106]], [[23, 22], [31, 50]]], dtype=torch.double
        )
        self.assertTensorsEqual(
            cplx.matmul(matrix1, matrix2),
            expect,
            msg="Matrix * Matrix multiplication failed!",
        )

    def test_scalar_inner_prod(self):
        scalar = torch.tensor([1, 2], dtype=torch.double)
        expect = torch.tensor([5, 0], dtype=torch.double)
        self.assertTensorsEqual(
            cplx.inner_prod(scalar, scalar), expect, msg="Scalar inner product failed!"
        )

    def test_vector_inner_prod(self):
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)
        expect = torch.tensor([30, 0], dtype=torch.double)
        self.assertTensorsEqual(
            cplx.inner_prod(vector, vector), expect, msg="Vector inner product failed!"
        )

    def test_outer_prod(self):
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)
        expect = torch.tensor(
            [[[10, 14], [14, 20]], [[0, 2], [-2, 0]]], dtype=torch.double
        )
        self.assertTensorsEqual(
            cplx.outer_prod(vector, vector), expect, msg="Outer product failed!"
        )

    def test_outer_prod_error_small(self):
        # take outer prod of 2 rank 1 tensors, instead of rank 2
        tensor = torch.tensor([1, 2], dtype=torch.double)
        with self.assertRaises(ValueError):
            cplx.outer_prod(tensor, tensor)

    def test_outer_prod_error_large(self):
        # take outer prod of 2 rank 3 tensors, instead of rank 2
        tensor = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)
        with self.assertRaises(ValueError):
            cplx.outer_prod(tensor, tensor)

    def test_conjugate(self):
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)

        expect = torch.tensor([[1, 2], [-3, -4]], dtype=torch.double)

        self.assertTensorsEqual(
            cplx.conjugate(vector), expect, msg="Vector conjugate failed!"
        )

    def test_matrix_conjugate(self):
        matrix = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)
        expect = torch.tensor(
            [[[1, 3], [2, 4]], [[-5, -7], [-6, -8]]], dtype=torch.double
        )
        self.assertTensorsEqual(
            cplx.conjugate(matrix), expect, msg="Matrix conjugate failed!"
        )

    def test_kronecker_prod(self):
        matrix = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)

        expect = torch.tensor(
            [
                [
                    [-24, -28, -28, -32],
                    [-32, -36, -36, -40],
                    [-32, -36, -36, -40],
                    [-40, -44, -44, -48],
                ],
                [
                    [10, 16, 16, 24],
                    [22, 28, 32, 40],
                    [22, 32, 28, 40],
                    [42, 52, 52, 64],
                ],
            ],
            dtype=torch.double,
        )

        self.assertTensorsEqual(
            cplx.kronecker_prod(matrix, matrix), expect, msg="Kronecker product failed!"
        )

    def test_kronecker_prod_error_small(self):
        # take KronProd of 2 rank 2 tensors, instead of rank 3
        tensor = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)
        with self.assertRaises(ValueError):
            cplx.kronecker_prod(tensor, tensor)

    def test_kronecker_prod_error_large(self):
        # take KronProd of 2 rank 4 tensors, instead of rank 3
        tensor = torch.arange(16, dtype=torch.double).reshape(2, 2, 2, 2)
        with self.assertRaises(ValueError):
            cplx.kronecker_prod(tensor, tensor)

    def test_vector_scalar_divide(self):
        scalar = torch.tensor([1, 2], dtype=torch.double)
        vector = torch.tensor([[1, 2], [3, 4]], dtype=torch.double)
        expect = torch.tensor([[1.4, 2.0], [0.2, 0.0]], dtype=torch.double)

        self.assertTensorsAlmostEqual(
            cplx.scalar_divide(vector, scalar),
            expect,
            msg="Vector / Scalar divide failed!",
        )

    def test_matrix_scalar_divide(self):
        scalar = torch.tensor([1, 2], dtype=torch.double)
        matrix = torch.tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=torch.double)

        expect = torch.tensor(
            [[[2.2, 2.8], [3.4, 4.0]], [[0.6, 0.4], [0.2, 0.0]]], dtype=torch.double
        )

        self.assertTensorsAlmostEqual(
            cplx.scalar_divide(matrix, scalar),
            expect,
            msg="Matrix / Scalar divide failed!",
        )

    def test_norm_sqr(self):
        scalar = torch.tensor([3, 4], dtype=torch.double)
        expect = torch.tensor(25, dtype=torch.double)

        self.assertTensorsEqual(cplx.norm_sqr(scalar), expect, msg="Norm failed!")

    def test_norm(self):
        scalar = torch.tensor([3, 4], dtype=torch.double)
        expect = torch.tensor(5, dtype=torch.double)

        self.assertTensorsEqual(cplx.norm(scalar), expect, msg="Norm failed!")

    def test_absolute_value(self):
        tensor = torch.tensor(
            [[[5, 5, -5, -5], [3, 6, -9, 1]], [[2, -2, 2, -2], [-7, 8, 0, 4]]],
            dtype=torch.double,
        )

        expect = torch.tensor(
            [[[np.sqrt(29)] * 4, [np.sqrt(58), 10, 9, np.sqrt(17)]]], dtype=torch.double
        )

        self.assertTensorsAlmostEqual(
            cplx.absolute_value(tensor), expect, msg="Absolute Value failed!"
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch

from qucumber.utils import unitaries


def test_default_unitary_dict():
    unitary_dict = unitaries.create_dict()

    msg = "Default Unitary dictionary has the wrong keys!"
    assert {"X", "Y", "Z"} == set(unitary_dict.keys()), msg


def test_adding_unitaries():
    unitary_dict = unitaries.create_dict(
        A=(
            0.5
            * torch.tensor([[[1, 1], [1, 1]], [[1, -1], [-1, 1]]], dtype=torch.double)
        )
    )

    msg = "Unitary dictionary has the wrong keys!"
    assert {"X", "Y", "Z", "A"} == set(unitary_dict.keys()), msg



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path
import pickle
from collections import namedtuple

import torch
import pytest

import qucumber
from qucumber.nn_states import PositiveWaveFunction, ComplexWaveFunction, DensityMatrix

from grads_utils import ComplexGradsUtils, PosGradsUtils, DensityGradsUtils
from conftest import all_state_types, assertAlmostEqual, TOL


SEED = 1234
EPS = 1e-6


def positive_wavefunction_data(request, gpu, num_hidden):
    with open(
        os.path.join(request.fspath.dirname, "data", "test_grad_data.pkl"), "rb"
    ) as f:
        test_data = pickle.load(f)

    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    data = torch.tensor(test_data["tfim1d"]["train_samples"], dtype=torch.double)
    target = torch.tensor(test_data["tfim1d"]["target_psi"], dtype=torch.double).t()

    num_visible = data.shape[-1]

    nn_state = PositiveWaveFunction(num_visible, num_hidden, gpu=gpu)
    PGU = PosGradsUtils(nn_state)

    data = data.to(device=nn_state.device)
    space = nn_state.generate_hilbert_space()
    target = target.to(device=nn_state.device)

    PositiveWaveFunctionFixture = namedtuple(
        "PositiveWaveFunctionFixture",
        ["data_samples", "target", "grad_utils", "nn_state", "space"],
    )

    return PositiveWaveFunctionFixture(
        data_samples=data, target=target, grad_utils=PGU, nn_state=nn_state, space=space
    )


def complex_wavefunction_data(request, gpu, num_hidden):
    with open(
        os.path.join(request.fspath.dirname, "data", "test_grad_data.pkl"), "rb"
    ) as f:
        test_data = pickle.load(f)

    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    data_bases = test_data["2qubits"]["train_bases"]
    data_samples = torch.tensor(
        test_data["2qubits"]["train_samples"], dtype=torch.double
    )

    all_bases = test_data["2qubits"]["bases"]
    target_psi_tmp = torch.tensor(
        test_data["2qubits"]["target_psi"], dtype=torch.double
    ).t()

    num_visible = data_samples.shape[-1]

    nn_state = ComplexWaveFunction(num_visible, num_hidden, gpu=gpu)
    unitary_dict = nn_state.unitary_dict

    CGU = ComplexGradsUtils(nn_state)

    all_bases = CGU.transform_bases(all_bases)

    target = CGU.load_target_psi(all_bases, target_psi_tmp)
    target = {b: v.to(device=nn_state.device) for b, v in target.items()}

    space = nn_state.generate_hilbert_space()
    data_samples = data_samples.to(device=nn_state.device)

    ComplexWaveFunctionFixture = namedtuple(
        "ComplexWaveFunctionFixture",
        [
            "data_samples",
            "data_bases",
            "grad_utils",
            "all_bases",
            "target",
            "space",
            "nn_state",
            "unitary_dict",
        ],
    )

    return ComplexWaveFunctionFixture(
        data_samples=data_samples,
        data_bases=data_bases,
        grad_utils=CGU,
        all_bases=all_bases,
        target=target,
        space=space,
        nn_state=nn_state,
        unitary_dict=unitary_dict,
    )


def density_matrix_data(request, gpu, num_hidden):
    with open(
        os.path.join(request.fspath.dirname, "data", "test_grad_data.pkl"), "rb"
    ) as f:
        test_data = pickle.load(f)

    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    data_bases = test_data["density_matrix"]["train_bases"]
    data_samples = torch.tensor(
        test_data["density_matrix"]["train_samples"], dtype=torch.double
    )

    all_bases = test_data["density_matrix"]["bases"]
    target = torch.tensor(
        test_data["density_matrix"]["density_matrix"], dtype=torch.double
    )

    num_visible = data_samples.shape[-1]
    num_aux = num_visible + 1

    nn_state = DensityMatrix(num_visible, num_hidden, num_aux, gpu=gpu)
    unitary_dict = nn_state.unitary_dict

    DGU = DensityGradsUtils(nn_state)

    all_bases = DGU.transform_bases(all_bases)

    space = nn_state.generate_hilbert_space()
    data_samples = data_samples.to(device=nn_state.device)
    target = target.to(device=nn_state.device)

    DensityMatrixFixture = namedtuple(
        "DensityMatrixFixture",
        [
            "data_samples",
            "data_bases",
            "grad_utils",
            "all_bases",
            "target",
            "space",
            "nn_state",
            "unitary_dict",
        ],
    )

    return DensityMatrixFixture(
        data_samples=data_samples,
        data_bases=data_bases,
        grad_utils=DGU,
        all_bases=all_bases,
        target=target,
        space=space,
        nn_state=nn_state,
        unitary_dict=unitary_dict,
    )


hidden_layer_sizes = [pytest.param(9, id="9", marks=[pytest.mark.extra]), 10]
grad_types = ["KL", "NLL"]


@pytest.fixture(scope="module", params=all_state_types)
def quantum_state_constructor(request):
    nn_state_type = request.param
    if nn_state_type == PositiveWaveFunction:
        return positive_wavefunction_data
    elif nn_state_type == ComplexWaveFunction:
        return complex_wavefunction_data
    elif nn_state_type == DensityMatrix:
        return density_matrix_data
    else:
        raise ValueError(
            f"invalid test config: {nn_state_type} is not a valid quantum state type"
        )


@pytest.fixture(scope="module", params=hidden_layer_sizes)
def quantum_state_data(request, quantum_state_constructor, quantum_state_device):
    return quantum_state_constructor(request, quantum_state_device, request.param)


@pytest.fixture(scope="module", params=grad_types)
def quantum_state_graddata(request, quantum_state_data):
    grad_type = request.param
    nn_state, grad_utils = quantum_state_data.nn_state, quantum_state_data.grad_utils

    if grad_type == "KL":
        alg_grad_fn = grad_utils.algorithmic_gradKL
        num_grad_fn = grad_utils.numeric_gradKL
    else:
        alg_grad_fn = grad_utils.algorithmic_gradNLL
        num_grad_fn = grad_utils.numeric_gradNLL

    alg_grads = alg_grad_fn(**quantum_state_data._asdict())
    num_grads = [None for _ in nn_state.networks]

    for n, net in enumerate(nn_state.networks):
        rbm = getattr(nn_state, net)
        num_grad = torch.tensor([]).to(device=rbm.device, dtype=torch.double)
        for param in rbm.parameters():
            num_grad = torch.cat(
                (
                    num_grad,
                    num_grad_fn(
                        param=param.view(-1), eps=EPS, **quantum_state_data._asdict()
                    ).to(num_grad),
                )
            )
        num_grads[n] = num_grad

    return nn_state, alg_grads, num_grads, grad_type, TOL


def get_param_status(i, param_ranges):
    """Get parameter name of the parameter in param_ranges which contains the index i.

    Also return whether i is pointing to the first index of the parameter.
    """
    for p, rng in param_ranges.items():
        if i in rng:
            return p, i == rng[0]


def test_grads(quantum_state_graddata):
    nn_state, alg_grads, num_grads, grad_type, test_tol = quantum_state_graddata

    print(
        "\nTesting {} gradients for {} on {}.".format(
            grad_type, nn_state.__class__.__name__, nn_state.device
        )
    )

    for n, net in enumerate(nn_state.networks):
        print("\nRBM: %s" % net)
        rbm = getattr(nn_state, net)

        param_ranges = {}
        counter = 0
        for param_name, param in rbm.named_parameters():
            param_ranges[param_name] = range(counter, counter + param.numel())
            counter += param.numel()

        for i, grad in enumerate(num_grads[n]):
            p_name, at_start = get_param_status(i, param_ranges)
            if at_start:
                print(f"\nTesting {p_name}...")
                print(f"Numerical {grad_type}\tAlg {grad_type}")

            print("{: 10.8f}\t{: 10.8f}\t\t".format(grad, alg_grads[n][i].item()))

        assertAlmostEqual(
            num_grads[n],
            alg_grads[n],
            test_tol,
            msg=f"{grad_type} grads are not close enough for {net}!",
        )



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from math import isclose

import pytest
import torch

from qucumber import observables, set_random_seed
from qucumber.nn_states import WaveFunctionBase, NeuralStateBase
from qucumber.utils import cplx, auto_unsqueeze_args

from conftest import all_state_types


SEED = 1234


class MockWaveFunction(WaveFunctionBase):
    _rbm_am = None
    _device = None

    def __init__(self, nqubits, state=None):
        self.nqubits = nqubits
        self.device = "cpu"

    def sample(self, num_samples=1):
        dist = torch.distributions.Bernoulli(probs=0.5)
        sample_size = torch.Size((num_samples, self.nqubits))
        initial_state = dist.sample(sample_size).to(
            device=self.device, dtype=torch.double
        )

        return initial_state

    @staticmethod
    def autoload(location, gpu=False):
        pass

    def gradient(self, v):
        return 0.0

    @property
    def rbm_am(self):
        return self._rbm_am

    @rbm_am.setter
    def rbm_am(self, new_val):
        self._rbm_am = new_val

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, new_val):
        self._device = new_val

    @property
    def networks(self):
        return ["rbm_am"]

    @auto_unsqueeze_args()
    def phase(self, v):
        return torch.zeros(v.shape[0], dtype=torch.double, device=self.device)

    def amplitude(self, v):
        return torch.ones(v.shape[0], dtype=torch.double, device=self.device)

    def psi(self, v):
        return cplx.make_complex(self.amplitude(v))


class MockDensityMatrix(NeuralStateBase):
    _rbm_am = None
    _device = None

    def __init__(self, nqubits, state=None):
        self.nqubits = nqubits
        self.device = "cpu"

    def sample(self, num_samples=1):
        dist = torch.distributions.Bernoulli(probs=0.5)
        sample_size = torch.Size((num_samples, self.nqubits))
        initial_state = dist.sample(sample_size).to(
            device=self.device, dtype=torch.double
        )

        return initial_state

    @staticmethod
    def autoload(location, gpu=False):
        pass

    def gradient(self, v):
        return 0.0

    @property
    def rbm_am(self):
        return self._rbm_am

    @rbm_am.setter
    def rbm_am(self, new_val):
        self._rbm_am = new_val

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, new_val):
        self._device = new_val

    @property
    def networks(self):
        return ["rbm_am"]

    def probability(self, v):
        return torch.ones(v.shape[0], dtype=torch.double, device=self.device)

    def importance_sampling_numerator(self, iter_sample, drawn_sample):
        out = torch.zeros(iter_sample.shape[0], dtype=torch.double, device=self.device)
        idx = torch.all(iter_sample == drawn_sample, dim=-1)
        out[idx] = self.probability(iter_sample[idx, :])
        return cplx.make_complex(out)

    def importance_sampling_denominator(self, drawn_sample):
        return cplx.make_complex(self.probability(drawn_sample))


@pytest.fixture(scope="module", params=[MockWaveFunction, MockDensityMatrix])
def mock_state_samples(request):
    set_random_seed(SEED, cpu=True, gpu=False, quiet=True)
    state_type = request.param

    pauli_expectations = {
        observables.SigmaX: (1.0, 1.0),
        observables.SigmaY: (0.0, 0.0),
        observables.SigmaZ: (0.0, 0.5),
    }

    if state_type == MockDensityMatrix:
        pauli_expectations[observables.SigmaX] = (0.0, 0.0)

    nn_state = state_type(2)
    test_sample = nn_state.sample(num_samples=100000)
    return nn_state, test_sample, pauli_expectations


def test_spinflip(mock_state_samples):
    _, test_sample, _ = mock_state_samples
    samples = test_sample.clone()
    observables.pauli.flip_spin(1, samples)  # flip spin
    observables.pauli.flip_spin(1, samples)  # flip it back
    assert torch.equal(samples, test_sample)


paulis = [
    pytest.param(observables.SigmaX, id="X"),
    pytest.param(observables.SigmaY, id="Y"),
    pytest.param(observables.SigmaZ, id="Z"),
]


@pytest.mark.parametrize("pauli", paulis)
@pytest.mark.parametrize(
    "absolute", [pytest.param(True, id="absolute"), pytest.param(False, id="signed")]
)
def test_pauli(mock_state_samples, pauli, absolute):
    nn_state, test_sample, pauli_expectations = mock_state_samples
    expectations = pauli_expectations[pauli]

    if absolute:
        mag = expectations[1]
        prefix = "(absolute) "
    else:
        mag = expectations[0]
        prefix = ""

    obs = pauli(absolute=absolute)
    measure_O = float(obs.apply(nn_state, test_sample).mean())

    assert isclose(
        mag, measure_O, abs_tol=1e-2
    ), "measure {}-magnetization failed".format(prefix + obs.symbol)


@pytest.fixture(scope="module", params=[2, 7])
def nn_state_num_visible(request):
    return request.param


@pytest.fixture(scope="module", params=all_state_types)
def nn_state(request, quantum_state_device, nn_state_num_visible):
    set_random_seed(SEED, cpu=True, gpu=quantum_state_device, quiet=True)
    state_type = request.param

    return state_type(nn_state_num_visible, gpu=quantum_state_device)


observable_constructors = [
    observables.SWAP,
    observables.SigmaX,
    observables.SigmaY,
    observables.SigmaZ,
    observables.NeighbourInteraction,
]

observable_constructors = [(cls.__name__, cls) for cls in observable_constructors]

observable_constructors.append(
    (
        "AlgebraicallyCombined",
        lambda: -observables.NeighbourInteraction() - (3 * observables.SigmaX()) + 1,
    )
)


@pytest.fixture(scope="module", params=observable_constructors, ids=lambda p: p[0])
def nn_state_observable_fixture(request, nn_state):
    obs_name, obs_constructor = request.param

    if obs_name == "SWAP":
        obs = obs_constructor(list(range(nn_state.num_visible // 2)))
    else:
        obs = obs_constructor()

    return nn_state, obs


@pytest.fixture(scope="module")
def nn_state_observable_samples(request, nn_state_observable_fixture):
    nn_state, obs = nn_state_observable_fixture

    test_sample = nn_state.sample(k=1, num_samples=1000)
    return nn_state, obs, test_sample


def test_sanity_check_observable_apply(request, nn_state_observable_samples):
    nn_state, obs, test_sample = nn_state_observable_samples
    applied = obs.apply(nn_state, test_sample)

    msg = "Output of `apply` should have the same number of samples as `test_sample`!"
    assert applied.shape[0] == test_sample.shape[0], msg


def test_sanity_check_observable_statistics_from_samples(
    request, nn_state_observable_samples
):
    nn_state, obs, test_sample = nn_state_observable_samples
    obs.statistics_from_samples(nn_state, test_sample)


def test_sanity_check_observable_statistics(request, nn_state_observable_fixture):
    nn_state, obs = nn_state_observable_fixture
    obs.statistics(nn_state, 1000)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager

import pytest

from qucumber.callbacks import (
    LambdaCallback,
    MetricEvaluator,
    ObservableEvaluator,
    EarlyStopping,
)


callback_stages = (
    "on_train_start",
    "on_train_end",
    "on_epoch_start",
    "on_epoch_end",
    "on_batch_start",
    "on_batch_end",
)


@pytest.mark.parametrize("stage", callback_stages)
def test_lambda_callback_value_error_num_args(stage):
    msg = f"LambdaCallback should fail if {stage} gets wrong # of arguments."

    with pytest.raises(ValueError):
        kwargs = {stage: lambda nn_state, epoch, batch, extra: "foobar"}
        LambdaCallback(**kwargs)
        pytest.fail(msg)


@pytest.mark.parametrize("stage", callback_stages)
def test_lambda_callback_type_error(stage):
    msg = f"LambdaCallback should fail if {stage} isn't a function or None."

    with pytest.raises(TypeError):
        LambdaCallback(**{stage: "foobar"})
        pytest.fail(msg)


@contextmanager
def no_exception():
    yield


es_params = [
    ("relative", no_exception()),
    ("absolute", no_exception()),
    ("variance", pytest.raises(TypeError)),
]


@pytest.mark.parametrize("criterion, exception", es_params)
def test_early_stopping_construction_metric(criterion, exception):
    ev = MetricEvaluator(1, {})
    with exception:
        EarlyStopping(1, 1, 1, ev, "", criterion=criterion)


@pytest.mark.parametrize("criterion", [crit for crit, exc in es_params])
def test_early_stopping_construction_observable(criterion):
    ev = ObservableEvaluator(1, [])
    EarlyStopping(1, 1, 1, ev, "", criterion=criterion)



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path
from functools import reduce
from operator import attrgetter

import torch
from torch.nn.utils import parameters_to_vector
import pytest

import qucumber
from qucumber.nn_states import PositiveWaveFunction, ComplexWaveFunction, DensityMatrix
from qucumber.utils import cplx
from qucumber.utils.unitaries import (
    create_dict,
    rotate_psi,
    rotate_rho,
    rotate_psi_inner_prod,
    rotate_rho_probs,
)

from conftest import devices, assertAlmostEqual, TOL


INIT_SEED = 1234  # seed to initialize model params with
SAMPLING_SEED = 1337  # seed to draw samples from the model with

all_state_types = [PositiveWaveFunction, ComplexWaveFunction, DensityMatrix]


@pytest.mark.parametrize("state_type", all_state_types)
def test_model_saving_and_loading(tmpdir, state_type):
    # some CUDA ops are non-deterministic; don't test on GPU.
    qucumber.set_random_seed(INIT_SEED, cpu=True, gpu=False, quiet=True)
    nn_state = state_type(10, gpu=False)

    model_path = str(tmpdir.mkdir("nn_state").join("params.pt").realpath())

    nn_state.save(model_path)

    qucumber.set_random_seed(SAMPLING_SEED, cpu=True, gpu=False, quiet=True)
    # don't worry about floating-point wonkyness
    orig_sample = nn_state.sample(k=10).to(dtype=torch.uint8)

    nn_state2 = state_type(10, gpu=False)

    nn_state2.load(model_path)

    qucumber.set_random_seed(SAMPLING_SEED, cpu=True, gpu=False, quiet=True)
    post_load_sample = nn_state2.sample(k=10).to(dtype=torch.uint8)

    msg = "Got different sample after reloading model!"
    assert torch.equal(orig_sample, post_load_sample), msg

    nn_state3 = state_type.autoload(model_path, gpu=False)

    qucumber.set_random_seed(SAMPLING_SEED, cpu=True, gpu=False, quiet=True)
    post_autoload_sample = nn_state3.sample(k=10).to(dtype=torch.uint8)

    msg = "Got different sample after autoloading model!"
    assert torch.equal(orig_sample, post_autoload_sample), msg

    os.remove(model_path)


@pytest.mark.parametrize("state_type", all_state_types)
@pytest.mark.parametrize("is_src_gpu", devices)
@pytest.mark.parametrize("is_dest_gpu", devices)
def test_autoloading(tmpdir, state_type, is_src_gpu, is_dest_gpu):
    model_path = str(tmpdir.mkdir("nn_state").join("params.pt").realpath())

    nn_state = state_type(10, gpu=is_src_gpu)
    nn_state.save(model_path)

    nn_state2 = state_type(10, gpu=is_dest_gpu)
    nn_state2.load(model_path)

    os.remove(model_path)


@pytest.mark.parametrize(
    "state_type, bad_key",
    [
        (PositiveWaveFunction, "rbm_am"),
        (ComplexWaveFunction, "rbm_am"),
        (ComplexWaveFunction, "rbm_ph"),
        (ComplexWaveFunction, "unitary_dict"),
        (DensityMatrix, "rbm_am"),
        (DensityMatrix, "rbm_ph"),
        (DensityMatrix, "unitary_dict"),
    ],
)
def test_model_saving_bad_metadata_key(tmpdir, state_type, bad_key):
    # some CUDA ops are non-deterministic; don't test on GPU.
    qucumber.set_random_seed(INIT_SEED, cpu=True, gpu=False, quiet=True)
    nn_state = state_type(10, gpu=False)

    model_path = str(tmpdir.mkdir("nn_state").join("params.pt").realpath())

    msg = "Metadata with invalid key should raise an error."
    with pytest.raises(ValueError):
        nn_state.save(model_path, metadata={bad_key: 1337})
        pytest.fail(msg)


def test_positive_wavefunction_phase():
    nn_state = PositiveWaveFunction(10, gpu=False)

    vis_state = torch.ones(10).to(dtype=torch.double)
    actual_phase = nn_state.phase(vis_state).to(vis_state)
    expected_phase = torch.zeros(1).to(vis_state).squeeze()

    msg = "PositiveWaveFunction is giving a non-zero phase for single visible state!"
    assert torch.equal(actual_phase, expected_phase), msg

    vis_state = torch.ones(10, 10).to(dtype=torch.double)
    actual_phase = nn_state.phase(vis_state).to(vis_state)
    expected_phase = torch.zeros(10).to(vis_state)

    msg = "PositiveWaveFunction is giving a non-zero phase for batch of visible states!"
    assert torch.equal(actual_phase, expected_phase), msg


def test_positive_wavefunction_psi():
    nn_state = PositiveWaveFunction(10, gpu=False)

    vis_state = torch.ones(10).to(dtype=torch.double)
    actual_psi_im = cplx.imag(nn_state.psi(vis_state)).to(vis_state)
    expected_psi_im = torch.zeros(1).squeeze().to(vis_state)

    msg = "PositiveWaveFunction is giving a non-zero imaginary part!"
    assert torch.equal(actual_psi_im, expected_psi_im), msg


def test_density_matrix_hermiticity():
    nn_state = DensityMatrix(5, 5, 5, gpu=False)

    space = nn_state.generate_hilbert_space(5)
    Z = nn_state.normalization(space)
    rho = nn_state.rho(space, space) / Z

    assert torch.equal(rho, cplx.conjugate(rho)), "DensityMatrix should be Hermitian!"


def test_density_matrix_tr1():
    nn_state = DensityMatrix(5, gpu=False)

    space = nn_state.generate_hilbert_space(5)
    matrix = nn_state.rho(space, space) / nn_state.normalization(space)

    msg = f"Trace of density matrix is not within {TOL} of 1!"
    assertAlmostEqual(torch.trace(matrix[0]), torch.Tensor([1]), TOL, msg=msg)


def test_density_matrix_sizes():
    nn_state = DensityMatrix(5, gpu=False)
    v = nn_state.generate_hilbert_space(5)
    vp = v[:4, :]

    rho = nn_state.rho(v, vp)

    assert rho.shape == (2, v.shape[0], vp.shape[0])


ndo_matrices = [
    ("rho", True),
    ("pi", True),
    ("rbm_am.gamma", False, +1),
    ("rbm_am.gamma", False, -1),
    ("pi_grad", True, True),
    ("pi_grad", True, False),
    ("rbm_am.gamma_grad", True, +1),
    ("rbm_am.gamma_grad", True, -1),
]


def _ndo_matrix_task_name(p):
    if len(p) < 3:
        return p[0]
    else:
        s = p[0]
        if p[2] is not True and p[2] == +1:
            s += "+"
        elif p[2] == -1:
            s += "-"
        else:
            s += "-" + str(p[2])
        return s


@pytest.mark.parametrize("prop", ndo_matrices, ids=_ndo_matrix_task_name)
def test_density_matrix_expansion(prop):
    qucumber.set_random_seed(INIT_SEED, cpu=True, gpu=False, quiet=True)

    nn_state = DensityMatrix(5, gpu=False)
    v = nn_state.generate_hilbert_space(5)
    vp = v[torch.randperm(v.shape[0]), :]

    prop_name = prop[0]
    is_complex = prop[1]
    args = prop[2:]
    fn = attrgetter(prop_name)(nn_state)

    matrix = fn(v, vp, *args, expand=True)
    diag = fn(v, vp, *args, expand=False)

    msg = f"Diagonal of matrix {prop_name} is wrong!"

    equation = "cii...->ci..." if is_complex else "ii->i"
    assertAlmostEqual(torch.einsum(equation, matrix), diag, TOL, msg=msg)


def test_density_matrix_diagonal():
    nn_state = DensityMatrix(5, gpu=False)
    v = nn_state.generate_hilbert_space(5)

    rho = nn_state.rho(v, expand=True)
    diag = nn_state.rho(v, expand=False)

    msg = "Diagonal of density matrix is wrong!"
    assertAlmostEqual(torch.einsum("cii...->ci...", rho), diag, TOL, msg=msg)


@pytest.mark.parametrize("state_type", all_state_types)
def test_single_sample(state_type):
    nn_state = state_type(10, 7, gpu=False)

    sample = nn_state.sample(k=10).squeeze()
    h_sample = nn_state.sample_h_given_v(sample)

    msg = "Single hidden sample should give a vector!"
    assert h_sample.dim() == 1, msg


@pytest.mark.parametrize("state_type", all_state_types)
def test_sampling_with_overwrite(state_type):
    nn_state = state_type(10, gpu=False)

    old_state = torch.empty(100, 10).bernoulli_().to(dtype=torch.double)
    initial_state = old_state.clone()

    sample = nn_state.sample(k=10, initial_state=initial_state, overwrite=True)

    assert torch.equal(sample, initial_state), "initial_state did not get overwritten!"
    assert not torch.equal(sample, old_state), "Markov Chain did not get updated!"


@pytest.mark.parametrize("state_type", all_state_types)
def test_bad_stop_training_val(state_type):
    nn_state = state_type(10, gpu=False)

    msg = "Setting stop_training to a non-boolean value should have raised an error."
    with pytest.raises(ValueError):
        nn_state.stop_training = "foobar"
        pytest.fail(msg)


@pytest.mark.parametrize("state_type", all_state_types)
def test_parameter_reinitialization(state_type):
    # some CUDA ops are non-deterministic; don't test on GPU.
    qucumber.set_random_seed(INIT_SEED, cpu=True, gpu=False, quiet=True)
    nn_state = state_type(10, gpu=False)

    old_params = parameters_to_vector(nn_state.rbm_am.parameters())
    nn_state.reinitialize_parameters()
    new_params = parameters_to_vector(nn_state.rbm_am.parameters())

    msg = "Model parameters did not get reinitialized!"
    assert not torch.equal(old_params, new_params), msg


@pytest.mark.parametrize("state_type", all_state_types)
def test_large_hilbert_space_fail(state_type):
    qucumber.set_random_seed(INIT_SEED, cpu=True, gpu=False, quiet=True)

    nn_state = state_type(10, gpu=False)
    max_size = nn_state.max_size

    msg = "Generating full Hilbert Space for more than {} qubits should fail.".format(
        max_size
    )
    with pytest.raises(ValueError):
        nn_state.generate_hilbert_space(size=max_size + 1)
        pytest.fail(msg)


@pytest.mark.parametrize("num_visible", [1, 2, 7])
@pytest.mark.parametrize("wvfn_type", [PositiveWaveFunction, ComplexWaveFunction])
def test_rotate_psi(num_visible, wvfn_type):
    nn_state = wvfn_type(num_visible, gpu=False)
    basis = "X" * num_visible
    unitary_dict = create_dict()

    space = nn_state.generate_hilbert_space()
    psi = nn_state.psi(space)

    psi_r_fast = rotate_psi(nn_state, basis, space, unitary_dict, psi=psi)

    U = reduce(cplx.kronecker_prod, [unitary_dict[b] for b in basis])
    psi_r_correct = cplx.matmul(U, psi)

    assertAlmostEqual(psi_r_fast, psi_r_correct, msg="Fast psi rotation failed!")


@pytest.mark.parametrize("num_visible", [1, 2, 7])
@pytest.mark.parametrize("state_type", [PositiveWaveFunction, ComplexWaveFunction])
@pytest.mark.parametrize(
    "precompute_psi",
    [pytest.param(True, id="precompute_psi"), pytest.param(False, id="psi_from_model")],
)
def test_rotate_psi_inner_prod(num_visible, state_type, precompute_psi):
    nn_state = state_type(num_visible, gpu=False)
    basis = "X" * num_visible
    unitary_dict = create_dict()

    space = nn_state.generate_hilbert_space()

    psi = nn_state.psi(space) if precompute_psi else None
    psi_r = rotate_psi(nn_state, basis, space, unitary_dict, psi=psi)

    psi_r_ip = rotate_psi_inner_prod(nn_state, basis, space, unitary_dict, psi=psi)

    assertAlmostEqual(psi_r, psi_r_ip, msg="Fast psi inner product rotation failed!")


@pytest.mark.parametrize("num_visible", [1, 2, 7])
@pytest.mark.parametrize("state_type", [DensityMatrix])
def test_rotate_rho(num_visible, state_type):
    nn_state = state_type(num_visible, gpu=False)
    basis = "X" * num_visible
    unitary_dict = create_dict()

    space = nn_state.generate_hilbert_space()
    rho = nn_state.rho(space, space)

    rho_r_fast = rotate_rho(nn_state, basis, space, unitary_dict, rho=rho)

    U = reduce(cplx.kronecker_prod, [unitary_dict[b] for b in basis])
    rho_r_correct = cplx.matmul(U, rho)
    rho_r_correct = cplx.matmul(rho_r_correct, cplx.conjugate(U))

    assertAlmostEqual(rho_r_fast, rho_r_correct, msg="Fast rho rotation failed!")


@pytest.mark.parametrize("num_visible", [1, 2, 7])
@pytest.mark.parametrize("state_type", [DensityMatrix])
@pytest.mark.parametrize(
    "precompute_rho",
    [pytest.param(True, id="precompute_rho"), pytest.param(False, id="rho_from_model")],
)
def test_rotate_rho_probs(num_visible, state_type, precompute_rho):
    nn_state = state_type(num_visible, gpu=False)
    basis = "X" * num_visible
    unitary_dict = create_dict()

    space = nn_state.generate_hilbert_space()

    rho = nn_state.rho(space, expand=True) if precompute_rho else None
    rho_r = rotate_rho(nn_state, basis, space, unitary_dict, rho=rho)
    rho_r_probs = torch.diagonal(cplx.real(rho_r))

    rho_r_probs_fast = rotate_rho_probs(nn_state, basis, space, unitary_dict, rho=rho)

    # use different tolerance as this sometimes just barely breaks through the
    #  smaller TOL value from test_grads.py
    assertAlmostEqual(
        rho_r_probs,
        rho_r_probs_fast,
        tol=(TOL * 10),
        msg="Fast rho probs rotation failed!",
    )



# Copyright 2019 PIQuIL - All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path

import pytest
import numpy as np
import torch
from torch.nn.utils import parameters_to_vector

import qucumber
import qucumber.utils.data as data
import qucumber.utils.training_statistics as ts
from qucumber.callbacks import LambdaCallback
from qucumber.nn_states import ComplexWaveFunction, PositiveWaveFunction, DensityMatrix

from conftest import devices, all_state_types

SEED = 1234


@pytest.mark.gpu
def test_complex_warn_on_gpu():
    with pytest.warns(ResourceWarning):
        ComplexWaveFunction(10, gpu=True)


@pytest.mark.parametrize("gpu", devices)
@pytest.mark.parametrize("state_type", all_state_types)
def test_neural_state(gpu, state_type):
    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)
    np.random.seed(SEED)

    nn_state = state_type(10, gpu=gpu)

    old_params = torch.cat(
        [
            parameters_to_vector(getattr(nn_state, net).parameters())
            for net in nn_state.networks
        ]
    )

    data = torch.ones(100, 10)

    # generate sample bases randomly, with probability 0.9 of being 'Z', otherwise 'X'
    bases = np.where(np.random.binomial(1, 0.9, size=(100, 10)), "Z", "X")

    nn_state.fit(data, epochs=1, pos_batch_size=10, input_bases=bases)

    new_params = torch.cat(
        [
            parameters_to_vector(getattr(nn_state, net).parameters())
            for net in nn_state.networks
        ]
    )

    msg = f"{state_type.__name__}'s parameters did not change!"
    assert not torch.equal(old_params, new_params), msg


def test_complex_training_without_bases_fail():
    qucumber.set_random_seed(SEED, cpu=True, gpu=False, quiet=True)

    nn_state = ComplexWaveFunction(10, gpu=False)

    data = torch.ones(100, 10)

    msg = "Training ComplexWaveFunction without providing bases should fail!"
    with pytest.raises(ValueError):
        nn_state.fit(data, epochs=1, pos_batch_size=10, input_bases=None)
        pytest.fail(msg)


@pytest.mark.parametrize("gpu", devices)
def test_stop_training(gpu):
    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    nn_state = PositiveWaveFunction(10, gpu=gpu)

    old_params = parameters_to_vector(nn_state.rbm_am.parameters())
    data = torch.ones(100, 10)

    nn_state.stop_training = True
    nn_state.fit(data)

    new_params = parameters_to_vector(nn_state.rbm_am.parameters())

    msg = "stop_training didn't work!"
    assert torch.equal(old_params, new_params), msg


def set_stop_training(nn_state):
    nn_state.stop_training = True


@pytest.mark.parametrize("gpu", devices)
def test_stop_training_in_batch(gpu):
    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    nn_state = PositiveWaveFunction(10, gpu=gpu)

    data = torch.ones(100, 10)

    callbacks = [
        LambdaCallback(on_batch_end=lambda nn_state, ep, b: set_stop_training(nn_state))
    ]

    nn_state.fit(data, callbacks=callbacks)

    msg = "stop_training wasn't set!"
    assert nn_state.stop_training, msg


@pytest.mark.parametrize("gpu", devices)
def test_stop_training_in_epoch(gpu):
    qucumber.set_random_seed(SEED, cpu=True, gpu=gpu, quiet=True)

    nn_state = PositiveWaveFunction(10, gpu=gpu)

    data = torch.ones(100, 10)

    callbacks = [
        LambdaCallback(on_epoch_end=lambda nn_state, ep: set_stop_training(nn_state))
    ]

    nn_state.fit(data, callbacks=callbacks)

    msg = "stop_training wasn't set!"
    assert nn_state.stop_training, msg


@pytest.fixture(scope="module", params=all_state_types)
def quantum_state_training_data(request):
    nn_state_type = request.param

    if nn_state_type == PositiveWaveFunction:

        root = os.path.join(
            request.fspath.dirname,
            "..",
            "examples",
            "Tutorial1_TrainPosRealWaveFunction",
        )

        train_samples, target = data.load_data(
            tr_samples_path=os.path.join(root, "tfim1d_data.txt"),
            tr_psi_path=os.path.join(root, "tfim1d_psi.txt"),
        )
        train_bases, bases = None, None

        nn_state = PositiveWaveFunction(num_visible=train_samples.shape[-1], gpu=False)

        batch_size, num_chains = 100, 200
        fid_target, kl_target = 0.85, 0.29

        reinit_params_fn = initialize_posreal_params

    elif nn_state_type == ComplexWaveFunction:

        root = os.path.join(
            request.fspath.dirname,
            "..",
            "examples",
            "Tutorial2_TrainComplexWaveFunction",
        )

        train_samples, target, train_bases, bases = data.load_data(
            tr_samples_path=os.path.join(root, "qubits_train.txt"),
            tr_psi_path=os.path.join(root, "qubits_psi.txt"),
            tr_bases_path=os.path.join(root, "qubits_train_bases.txt"),
            bases_path=os.path.join(root, "qubits_bases.txt"),
        )

        nn_state = ComplexWaveFunction(num_visible=train_samples.shape[-1], gpu=False)

        batch_size, num_chains = 50, 10
        fid_target, kl_target = 0.38, 0.33

        reinit_params_fn = initialize_complex_params

    elif nn_state_type == DensityMatrix:

        root = os.path.join(
            request.fspath.dirname, "..", "examples", "Tutorial3_TrainDensityMatrix"
        )

        train_samples, target, train_bases, bases = data.load_data_DM(
            tr_samples_path=os.path.join(root, "N2_W_state_100_samples_data.txt"),
            tr_mtx_real_path=os.path.join(root, "N2_W_state_target_real.txt"),
            tr_mtx_imag_path=os.path.join(root, "N2_W_state_target_imag.txt"),
            tr_bases_path=os.path.join(root, "N2_W_state_100_samples_bases.txt"),
            bases_path=os.path.join(root, "N2_IC_bases.txt"),
        )

        nn_state = DensityMatrix(num_visible=train_samples.shape[-1], gpu=False)

        batch_size, num_chains = 100, 10
        fid_target, kl_target = 0.45, 0.42

        def reinit_params_fn(request, nn_state):
            nn_state.reinitialize_parameters()

    else:
        raise ValueError(
            f"invalid test config: {nn_state_type} is not a valid quantum state type"
        )

    return {
        "nn_state": nn_state,
        "data": train_samples,
        "input_bases": train_bases,
        "target": target,
        "bases": bases,
        "epochs": 5,
        "pos_batch_size": batch_size,
        "neg_batch_size": num_chains,
        "k": 10,
        "lr": 0.1,
        "space": nn_state.generate_hilbert_space(),
        "fid_target": fid_target,
        "kl_target": kl_target,
        "reinit_params_fn": reinit_params_fn,
    }


@pytest.mark.slow
def test_training(request, quantum_state_training_data):
    qucumber.set_random_seed(SEED, cpu=True, gpu=False, quiet=True)

    fidelities = []
    KLs = []

    qstd = quantum_state_training_data

    print("Training 10 times and checking fidelity and KL at 5 epochs...\n")
    for i in range(10):
        print(f"Iteration: {i + 1}")

        qstd["reinit_params_fn"](request, qstd["nn_state"])

        qstd["nn_state"].fit(time=True, progbar=False, **qstd)

        fidelities.append(ts.fidelity(**qstd))
        KLs.append(ts.KL(**qstd))
        print(f"Fidelity: {fidelities[-1]}; KL: {KLs[-1]}.")

    print("\nStatistics")
    print("----------")
    print(
        "Fidelity: ",
        np.average(fidelities),
        "+/-",
        np.std(fidelities) / np.sqrt(len(fidelities)),
        "\n",
    )
    print("KL: ", np.average(KLs), "+/-", np.std(KLs) / np.sqrt(len(KLs)), "\n")

    assert abs(np.average(fidelities) - qstd["fid_target"]) < 0.02
    assert abs(np.average(KLs) - qstd["kl_target"]) < 0.02
    assert (np.std(fidelities) / np.sqrt(len(fidelities))) < 0.02
    assert (np.std(KLs) / np.sqrt(len(KLs))) < 0.02


def initialize_posreal_params(request, nn_state):
    with open(
        os.path.join(
            request.fspath.dirname, "data", "test_training_init_pos_params.npz"
        ),
        "rb",
    ) as f:
        x = np.load(f)
        for p in x.files:
            getattr(nn_state.rbm_am, p).data = torch.tensor(x[p]).to(
                getattr(nn_state.rbm_am, p)
            )


def initialize_complex_params(request, nn_state):
    with open(
        os.path.join(
            request.fspath.dirname, "data", "test_training_init_complex_params.npz"
        ),
        "rb",
    ) as f:
        x = np.load(f)
        for p in x.files:
            if p.startswith("am"):
                rbm = nn_state.rbm_am
            elif p.startswith("ph"):
                rbm = nn_state.rbm_ph

            q = p.split("_", maxsplit=1)[-1]
            getattr(rbm, q).data = torch.tensor(x[p]).to(getattr(rbm, q))
